# -- coding: ISO-8859-2 --

import requests, json, os, traceback, configparser, re, base64
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from typing import Literal


class BaarLocker:
    """This class contains all the functionalities to interact with BAARLocker APIs and performs operations like -
    Encrypt, Decrypt, Update and Delete"""
    def __init__(self, tenant: str, api_key: str, username: str, env_var: str ='BAAR_PATH', custom_url_dict: dict = {},
                 baar_version: Literal['BaarStudio', 'BaarStandalone', ''] = 'BaarStudio', secret_key: bytes = b'',
                 baar_config_file_path = ''):
        self.tenant = tenant
        self.api_key = api_key
        self.username = username
        self.env_var = env_var
        self.custom_url_dict = custom_url_dict
        self.baar_version = baar_version
        self.baar_config_file_path = baar_config_file_path
        match baar_version:
            case 'BaarStudio':
                server_type = 'baar-ws'
            case 'BaarStandalone':
                server_type = 'baar-locker'
            case _:
                server_type = 'baar-ws'
                print(f'Invalid BAAR Version provided "{baar_version}. Supported values: BaarStudio, BaarStandalone"')
                print(f'Using default BAAR Version BaarStudio')
        self.url_dict = {
            'update_url': f'__BASE_URL__/{server_type}/rest/{username}/UpdateBaarLocker/1?tenant={tenant}',
            'create_url': f'__BASE_URL__/{server_type}/rest/{username}/CreateBaarLocker/1?tenant={tenant}',
            'get_url': f'__BASE_URL__/{server_type}/rest/{username}/GetBaarLocker/1?tenant={tenant}',
            'delete_url': f'__BASE_URL__/{server_type}/rest/{username}/DeleteBaarLocker/1/?tenant={tenant}'
        }
        self.required_key_set = {'create_url', 'get_url', 'update_url', 'delete_url'}
        # Secret key (32 bytes for AES-256), default key for BaarLocker API Encryption/ Decryption
        if secret_key == b'':
            self.secret_key = b"12345678901234567890123456789012"
        else:
            self.secret_key = secret_key

    def smart_pre_post_processing(self, operation_type: Literal['encrypt', 'decrypt', 'update', 'delete'], data: str = ''):
        final_data = ''
        match self.baar_version:
            case 'BaarStudio':
                match operation_type:
                    case 'encrypt':
                        final_data = data
                    case 'decrypt':
                        final_data = data
                    case 'update':
                        final_data = data
                    case 'delete':
                        final_data = data
                    case _:
                        print(f'Invalid operation type provided - "{operation_type}"')
            case 'BaarStandalone':
                match operation_type:
                    case 'encrypt':
                        final_data = self.baar_api_payload_encryptor(data)
                    case 'decrypt':
                        final_data = self.baar_api_payload_decrypt(data)
                    case 'update':
                        final_data = self.baar_api_payload_encryptor(data)
                    case 'delete':
                        final_data = data
                    case _:
                        print(f'Invalid operation type provided - "{operation_type}"')
            case _:
                print(f'Invalid BAAR Version provided - "{operation_type}"')
        return final_data

    def baar_api_payload_encryptor(self, plaintext):
        # Encrypt function
        try:
            cipher = AES.new(self.secret_key, AES.MODE_ECB)
            padded_data = pad(plaintext.encode(), AES.block_size)  # PKCS5/PKCS7 padding
            encrypted_bytes = cipher.encrypt(padded_data)
            return base64.b64encode(encrypted_bytes).decode()  # Base64 encoded string
        except Exception as e:
            print('BAAR API Encryption Failed!')
            print(traceback.format_exc())
            return plaintext

    def baar_api_payload_decrypt(self, encrypted_text):
        # Decrypt function
        try:
            cipher = AES.new(self.secret_key, AES.MODE_ECB)
            encrypted_bytes = base64.b64decode(encrypted_text)
            decrypted_data = unpad(cipher.decrypt(encrypted_bytes), AES.block_size)  # Remove padding
            return decrypted_data.decode()  # Decoded UTF-8 string
        except Exception as e:
            print('BAAR API Decryption Failed!')
            print(traceback.format_exc())
            return encrypted_text

    def get_specific_url(self, url_type):
        # This function will filter the exact url based on the url_type from the self.url_dict
        final_url = ''
        if url_type in self.url_dict.keys():
            if len(self.custom_url_dict) == 0:
                base_url = self.get_base_url()
                for key in self.url_dict:
                    self.url_dict[key] = self.url_dict.get(key, '').replace('__BASE_URL__', base_url)
                main_url_dict = self.url_dict
                final_url = main_url_dict.get(url_type, '')
            else:
                if url_type in self.custom_url_dict.keys():
                    main_url_dict = self.custom_url_dict
                    final_url = main_url_dict.get(url_type, '')
                else:
                    print(f'''Configuration: custom_url_dict is missing required field "{url_type}"
                    to perform the required operation.''', flush=True)
                    print('''Following are the required field list: create_url, get_url, update_url, delete_url''',
                          flush=True)
        else:
            print(f'''Configuration: custom_url_dict is missing required field "{url_type}"
            to perform the required operation.''', flush=True)
            print('''Following are the required field list: create_url, get_url, update_url, delete_url''',
                  flush=True)
        return final_url

    def path_fetcher(self):
        # This function will fetch the Baar installation path from environment variables
        file_path = os.getenv(self.env_var)
        return file_path

    @staticmethod
    def config_reader(path, key_list: list):
        key_value_dict = {}
        try:
            if os.path.isfile(path):
                with open(path, 'r') as text_file:
                    for line in text_file:
                        if str(line).strip() != '' and str(line).strip() != '\n':
                            for key in key_list:
                                if key in line:
                                    value = line.split(key)[1].replace('\n', '')
                                    key_value_dict[key] = value
                                    if len((set(key_list).difference(set(key_value_dict.keys())))) == 0:
                                        break
                            if len((set(key_list).difference(set(key_value_dict.keys())))) == 0:
                                break
            else:
                print(f'Config file "{path}" does not exist!', flush=True)
        except:
            print(f'Unable to read/ extract data from config file "{path}"', flush=True)
        return key_value_dict

    @staticmethod
    def extract_base_url(full_url: str)-> str:
        base_url_regex = r"^(https?:\/\/[^\/]+)"
        base_url_matches = re.findall(base_url_regex, full_url, re.MULTILINE)
        if base_url_matches:
            return base_url_matches[0]
        else:
            print('Unable to find Base URL for BAAR Services.')
            return ''

    def get_base_url(self):
        # This function will automatically fetch the service url from the BAAR Properties files
        base_url = ''
        if str(self.baar_config_file_path).strip() != '':
            path = self.baar_config_file_path
        else:
            path = self.path_fetcher()
        if os.path.isdir(path):
            config_path = os.path.join(path, r'studio\config\studio-web.properties')
            if os.path.isfile(config_path):
                service_url = self.config_reader(config_path, ['service.url='])
                if len(service_url) > 0:
                    base_url = self.extract_base_url(service_url['service.url='])
                else:
                    base_url = ''
                if base_url == '':
                    print('Base URL Not Found!', flush=True)
            else:
                print('Config File does not exist.', flush=True)
        else:
            print(f'BAAR Studio Installation Path not found - {path}', flush=True)
        return base_url

    @staticmethod
    def verify_response(response):
        # This function will verify the responses received from API Communication
        good_response = False
        response_dict = {}
        if response:
            response_dict = json.loads(response.text)
            status_code = response_dict.get('statusCode', 0)
            message = response_dict.get('message', '')
            if status_code == 200 and 'successful' in str(message).lower().strip():
                good_response = True
            else:
                good_response = False
                print('Unable to connect to server!', flush=False)
        else:
            print('Unable to connect to server!', flush=False)
        return good_response, response_dict

    def encrypt(self, data: str):
        # This function will encrypt the provided data using BAARLocker API
        url = self.get_specific_url('create_url')
        final_unique_id = ''
        detailed_message = ''
        operation_status = False
        payload = json.dumps({"password": self.smart_pre_post_processing(operation_type = 'encrypt', data = data)})
        headers = {
            'username': self.username,
            'API-KEY': self.api_key,
            'X-NO-LOGIN-MODE': '',
            'Content-Type': 'application/json'
        }
        try:
            response = requests.request("POST", url, headers=headers, data=payload)
            good_response, response_dict = self.verify_response(response)
            if response and good_response:
                detailed_message = response_dict.get('message', '')
                unique_id = response_dict.get('uniqueId', '')
                if unique_id != '':
                    final_unique_id = unique_id
                    operation_status = True
                else:
                    print('Encryption Failed. Unique ID is empty!', flush=True)
                    detailed_message = response.text
            else:
                print('Encryption Failed!', flush=True)
                detailed_message = response.text
        except Exception as e:
            print('Encryption Failed!', flush=True)
            print(traceback.format_exc(), flush=True)
            detailed_message = 'Unable to connect to server and execute operation.'
        return final_unique_id, operation_status, detailed_message

    def decrypt(self, unique_id):
        # This function will decrypt the provided id using BAARLocker API
        url = self.get_specific_url('get_url')
        final_data = ''
        detailed_message = ''
        operation_status = False
        payload = json.dumps({"uniqueId": unique_id})
        headers = {
            'username': self.username,
            'API-KEY': self.api_key,
            'X-NO-LOGIN-MODE': '',
            'Content-Type': 'application/json'
        }
        try:
            response = requests.request("GET", url, headers=headers, data=payload)
            good_response, response_dict = self.verify_response(response)
            if response and good_response:
                detailed_message = response_dict.get('message', '')
                decrypted_data = response_dict.get('password', '')
                if decrypted_data != '':
                    final_data = self.smart_pre_post_processing(operation_type='decrypt', data=decrypted_data)
                    operation_status = True
                else:
                    print('Decryption Failed. Decrypted data is empty!', flush=True)
                    detailed_message = response.text
            else:
                print('Decryption Failed!', flush=True)
                detailed_message = response.text
        except Exception as e:
            print('Decryption Failed!', flush=True)
            print(traceback.format_exc(), flush=True)
            detailed_message = 'Unable to connect to server and execute decryption operation.'
        return final_data, operation_status, detailed_message

    def update(self, unique_id, data):
        # This function will Update the provided data against the provided if using BAARLocker API
        url = self.get_specific_url('update_url')
        final_unique_id = ''
        detailed_message = ''
        operation_status = False
        payload = json.dumps({
            "uniqueId": unique_id,
            "password": self.smart_pre_post_processing(operation_type = 'update', data = data)
        })
        headers = {
            'username': self.username,
            'API-KEY': self.api_key,
            'X-NO-LOGIN-MODE': '',
            'Content-Type': 'application/json'
        }
        try:
            response = requests.request("PUT", url, headers=headers, data=payload)
            good_response, response_dict = self.verify_response(response)
            if response and good_response:
                detailed_message = response_dict.get('message', '')
                unique_id = response_dict.get('uniqueId', '')
                if unique_id != '':
                    final_unique_id = unique_id
                    operation_status = True
                else:
                    print('Update Failed. Unique ID is empty!', flush=True)
                    detailed_message = response.text
            else:
                print('Update Failed!', flush=True)
                detailed_message = response.text
        except Exception as e:
            print('Update Failed!', flush=True)
            print(traceback.format_exc(), flush=True)
            detailed_message = 'Unable to connect to server and execute update operation.'
        return final_unique_id, operation_status, detailed_message

    def delete(self, unique_id):
        # This function will delete the provided id using BAARLocker API
        url = self.get_specific_url('delete_url')
        detailed_message = ''
        final_unique_id = unique_id
        operation_status = False
        payload = json.dumps({"uniqueId": unique_id})
        headers = {
            'username': self.username,
            'API-KEY': self.api_key,
            'X-NO-LOGIN-MODE': '',
            'Content-Type': 'application/json'
        }
        try:
            response = requests.request("DELETE", url, headers=headers, data=payload)
            good_response, response_dict = self.verify_response(response)
            if response and good_response:
                detailed_message = response_dict.get('message', '')
                operation_status = True
            else:
                print('Deletion Failed!', flush=True)
                detailed_message = response.text
        except Exception as e:
            print('Deletion Failed!', flush=True)
            print(traceback.format_exc(), flush=True)
            detailed_message = 'Unable to connect to server and execute deletion operation.'
        return final_unique_id, operation_status, detailed_message
